from pydantic import BaseSettings, Field

_FORMAT_WITH_TIME = (
    "<level>{level:<10}</level>|"
    "<yellow>{time:YYYY-MM-DD HH:mm:ss}</yellow>|"
    "<magenta>{extra}</magenta>|"
    "<blue>{function}</blue>|"
    "<level>{message}</level>"
)
_FORMAT_WITHOUT_TIME = (
    "<level>{level:<10}</level>|"
    "<magenta>{extra}</magenta>|"
    "<blue>{function}</blue>|"
    "<level>{message}</level>"
)


class LoggerConfig(BaseSettings):
    no_color: bool = Field(False, env="LOGGER_NO_COLOR")
    fmt_without_time: bool = Field(False, env="LOGGER_NO_TIME")

    @property
    def fmt(self) -> str:
        if self.fmt_without_time:
            return _FORMAT_WITHOUT_TIME
        return _FORMAT_WITH_TIME



from .config import LoggerConfig


def build_logger():
    import sys

    from loguru import logger

    cfg = LoggerConfig()
    logger.remove()

    logger.add(sys.stderr, format=cfg.fmt, colorize=not cfg.no_color)
    return logger


logger = build_logger()



import pandas as pd
from ..datamodel import Document
import typing as t
from sklearn.metrics.pairwise import cosine_similarity


def vector_search(
    docs: t.List[Document], model_hf, user_input: str = "I love you"
) -> pd.DataFrame:
    df_docs = pd.DataFrame.from_records([doc.dict() for doc in docs])
    df_emb = df_docs[["doc_id", "embedding"]]
    df_emb = df_emb.set_index("doc_id")
    df_emb = pd.DataFrame(list(df_emb["embedding"]))
    df_emb.index = df_docs["doc_id"]

    query_embedding = model_hf.embed_documents([user_input])

    df_query = pd.DataFrame(query_embedding)
    df_query.index = ["query_id"]

    # Compute the Cosine Similarity
    full_emb = pd.concat([df_emb, df_query])
    df_bert = pd.DataFrame(cosine_similarity(full_emb))

    df_bert.index = full_emb.index
    df_bert.columns = full_emb.index

    df_bert = df_bert.iloc[-1:,].T
    df_bert = df_bert.sort_values("query_id", ascending=False).reset_index()
    df_bert = df_bert[1:]

    df_bert = df_bert.rename(columns={"index": "doc_id"})
    final_df = pd.merge(df_bert, df_docs[["doc_id", "content"]], on="doc_id")
    final_df = final_df.rename(columns={"query_id": "cosine_similarity_score"})

    return final_df



from bunkatopics.datamodel import Document, ContinuumDimension, BourdieuDimension
from bunkatopics.visualisation.visu_utils import wrap_by_word
import typing as t
import pandas as pd
from sklearn.preprocessing import MinMaxScaler
from sklearn.metrics.pairwise import cosine_similarity
import plotly.express as px
from langchain.embeddings import HuggingFaceInstructEmbeddings
from bunkatopics.functions.topics import topic_modeling
from bunkatopics.functions.topic_representation import remove_overlapping_terms


def get_continuum(
    model_hf: HuggingFaceInstructEmbeddings,
    docs: t.List[Document],
    cont_name="emotion",
    left_words=["hate", "pain"],
    right_words=["love", "good"],
) -> pd.DataFrame:
    df_docs = pd.DataFrame.from_records([doc.dict() for doc in docs])
    df_emb = df_docs[["doc_id", "embedding"]]
    df_emb = df_emb.set_index("doc_id")
    df_emb = pd.DataFrame(list(df_emb["embedding"]))
    df_emb.index = df_docs["doc_id"]

    continuum = ContinuumDimension(
        id=cont_name, left_words=left_words, right_words=right_words
    )

    # Compute the extremity embeddings
    left_embedding = model_hf.embed_documents(continuum.left_words)
    right_embedding = model_hf.embed_documents(continuum.right_words)

    left_embedding = pd.DataFrame(left_embedding).mean().values.reshape(1, -1)
    right_embedding = pd.DataFrame(right_embedding).mean().values.reshape(1, -1)

    # Make the difference to get the continnum
    continuum_embedding = left_embedding - right_embedding
    df_continuum = pd.DataFrame(continuum_embedding)
    df_continuum.index = ["distance"]

    # Compute the Cosine Similarity
    full_emb = pd.concat([df_emb, df_continuum])
    df_bert = pd.DataFrame(cosine_similarity(full_emb))

    df_bert.index = full_emb.index
    df_bert.columns = full_emb.index
    df_bert = df_bert.iloc[-1:,].T
    df_bert = df_bert.sort_values("distance", ascending=False).reset_index()
    df_bert = df_bert[1:]
    df_bert = df_bert.rename(columns={"index": "doc_id"})
    final_df = pd.merge(df_bert, df_docs[["doc_id", "content"]], on="doc_id")

    scaler = MinMaxScaler(feature_range=(-1, 1))
    final_df[["distance"]] = scaler.fit_transform(final_df[["distance"]])

    final_df = final_df.set_index("doc_id")
    final_df = final_df[["distance"]]

    distance_dict = final_df.to_dict("index")

    for doc in docs:
        res = BourdieuDimension(
            continuum=continuum, distance=distance_dict.get(doc.doc_id)["distance"]
        )
        doc.bourdieu_dimensions.append(res)

    return docs


def visualize_bourdieu(
    model_hf,
    docs: t.List[Document],
    x_left_words: t.List[str] = ["war"],
    x_right_words: t.List[str] = ["peace"],
    y_top_words: t.List[str] = ["men"],
    y_bottom_words: t.List[str] = ["women"],
    height=1500,
    width=1500,
    clustering=True,
    n_clusters=5,
):
    # Reset
    for doc in docs:
        doc.bourdieu_dimensions = []

    # Compute Continuums
    new_docs = get_continuum(
        model_hf,
        docs,
        cont_name="cont1",
        left_words=x_left_words,
        right_words=x_right_words,
    )
    new_docs = get_continuum(
        model_hf,
        docs,
        cont_name="cont2",
        left_words=y_top_words,
        right_words=y_bottom_words,
    )

    df_names = [
        {
            "names": [y.continuum.id for y in x.bourdieu_dimensions],
            "left_words": [y.continuum.left_words for y in x.bourdieu_dimensions],
            "right_words": [y.continuum.right_words for y in x.bourdieu_dimensions],
        }
        for x in new_docs
    ]
    df_names = pd.DataFrame(df_names)
    df_names = df_names.explode(["names", "left_words", "right_words"])
    df_names["left_words"] = df_names["left_words"].apply(lambda x: "-".join(x))
    df_names["right_words"] = df_names["right_words"].apply(lambda x: "-".join(x))
    df_names = df_names.drop_duplicates()
    df_names = df_names.set_index("names")

    dict_bourdieu = df_names.to_dict(orient="index")

    df_bourdieu = [
        {
            "doc_id": x.doc_id,
            "coordinates": [y.distance for y in x.bourdieu_dimensions],
            "names": [y.continuum.id for y in x.bourdieu_dimensions],
        }
        for x in new_docs
    ]

    df_content = [{"doc_id": x.doc_id, "content": x.content} for x in new_docs]
    df_content = pd.DataFrame(df_content)

    df_bourdieu = pd.DataFrame(df_bourdieu)
    df_bourdieu = df_bourdieu.explode(["coordinates", "names"])

    df_fig = df_bourdieu[["doc_id", "coordinates", "names"]]
    df_fig = df_fig.pivot(index="doc_id", columns="names", values="coordinates")
    df_fig = df_fig.reset_index()
    df_fig = pd.merge(df_content, df_fig, on="doc_id")
    df_fig["content_plotly"] = df_fig["content"].apply(lambda x: wrap_by_word(x, 10))

    x_axis_name = list(dict_bourdieu.keys())[0]
    y_axis_name = list(dict_bourdieu.keys())[1]

    x_left_words = dict_bourdieu[x_axis_name]["left_words"]
    x_right_words = dict_bourdieu[x_axis_name]["right_words"]
    y_top_words = dict_bourdieu[y_axis_name]["left_words"]
    y_bottom_words = dict_bourdieu[y_axis_name]["right_words"]

    fig = px.density_contour(
        df_fig,
        x=x_axis_name,
        y=y_axis_name,
        # hover_data=["content_plotly"],
        template="simple_white",
        height=height,
        width=width,
        marginal_x="histogram",
        marginal_y="histogram",
        color_discrete_sequence=["grey"],
    )

    fig2 = px.scatter(
        df_fig,
        x=x_axis_name,
        y=y_axis_name,
        hover_data=["content_plotly"],
        template="simple_white",
        height=height,
        width=width,
        opacity=0.2,
        title="Bourdieu Plot",
        color_discrete_sequence=["blue"],
    )
    fig.add_traces(fig2.data)
    fig.update_xaxes(showgrid=False, showticklabels=False, zeroline=True)
    fig.update_yaxes(showgrid=False, showticklabels=False, zeroline=True)

    fig.update_layout(
        annotations=[
            dict(
                x=0,
                y=1,
                xref="x",
                yref="y",
                text=y_top_words,
                showarrow=False,
                xanchor="right",
                yanchor="top",
            ),
            dict(
                x=0,
                y=-1,
                xref="x",
                yref="y",
                text=y_bottom_words,
                showarrow=False,
                xanchor="left",
                yanchor="bottom",
            ),
            dict(
                x=1,
                y=0,
                xref="x",
                yref="y",
                text=x_left_words,
                showarrow=False,
                xanchor="right",
                yanchor="top",
            ),
            dict(
                x=-1,
                y=0,
                xref="x",
                yref="y",
                text=x_right_words,
                showarrow=False,
                xanchor="left",
                yanchor="bottom",
            ),
        ]
    )

    if clustering:
        # Get the Topic Modeling
        doc_ids = list(df_fig["doc_id"])
        x = df_fig[x_axis_name]
        y = df_fig[y_axis_name]
        docs_terms = [doc.term_id for doc in docs]

        dict_doc_embeddings = {
            doc_id: {"x": x_val, "y": y_val}
            for doc_id, x_val, y_val in zip(doc_ids, x, y)
        }
        dict_doc_terms = {
            doc_id: {"term_id": term} for doc_id, term in zip(doc_ids, docs_terms)
        }
        dict_topics, dict_docs = topic_modeling(
            dict_doc_embeddings,
            dict_doc_terms,
            n_clusters=n_clusters,
        )

        df_topics = pd.DataFrame(dict_topics).T
        df_topics["name"] = df_topics["term_id"].apply(lambda x: x[:5])
        df_topics["name"] = df_topics["name"].apply(
            lambda x: remove_overlapping_terms(x)
        )
        df_topics["name"] = df_topics["name"].apply(lambda x: " | ".join(x))
        df_topics["name_plotly"] = df_topics["name"].apply(lambda x: wrap_by_word(x, 7))

        topics_x = list(df_topics["x_centroid"])
        topics_y = list(df_topics["y_centroid"])
        topics_name_plotly = list(df_topics["name_plotly"])
        label_size_ratio = 100

        for x, y, label in zip(topics_x, topics_y, topics_name_plotly):
            fig.add_annotation(
                x=x,
                y=y,
                text=label,
                font=dict(
                    family="Courier New, monospace",
                    size=width / label_size_ratio,
                    color="red",
                ),
                bordercolor="#c7c7c7",
                borderwidth=width / 1000,
                borderpad=width / 500,
                bgcolor="white",
                opacity=1,
            )

    return fig



import numpy as np
from scipy import interpolate
from scipy.spatial import ConvexHull


def get_convex_hull_coord(points: np.array, interpolate_curve: bool = True):
    hull = ConvexHull(points)
    x_hull = np.append(points[hull.vertices, 0], points[hull.vertices, 0][0])
    y_hull = np.append(points[hull.vertices, 1], points[hull.vertices, 1][0])

    if interpolate_curve:
        # interpolate
        dist = np.sqrt(
            (x_hull[:-1] - x_hull[1:]) ** 2 + (y_hull[:-1] - y_hull[1:]) ** 2
        )
        dist_along = np.concatenate(([0], dist.cumsum()))
        spline, u = interpolate.splprep([x_hull, y_hull], u=dist_along, s=0, per=1)
        interp_d = np.linspace(dist_along[0], dist_along[-1], 50)
        interp_x, interp_y = interpolate.splev(interp_d, spline)
    else:
        interp_x = x_hull
        interp_y = y_hull

    return interp_x, interp_y



import typing as t
from pydantic import BaseModel, Field

TOPIC_ID = str
TERM_ID = str
DOC_ID = str


class ContinuumDimension(BaseModel):
    id: str = "continuum"
    left_words: t.List[str] = Field(default_factory=list)
    right_words: t.List[str] = Field(default_factory=list)


class BourdieuDimension(BaseModel):
    continuum: ContinuumDimension
    distance: float


class Document(BaseModel):
    doc_id: DOC_ID
    content: str
    size: t.Optional[float] = None
    x: t.Optional[float] = None
    y: t.Optional[float] = None
    topic_id: t.Optional[TOPIC_ID] = None
    term_id: t.Optional[t.List[TERM_ID]] = None
    embedding: t.Optional[t.List[float]] = Field(None, repr=False)
    bourdieu_dimensions: t.List[BourdieuDimension] = []


class ConvexHullModel(BaseModel):
    topic_id: TOPIC_ID
    x_coordinates: t.Optional[t.List[float]] = None
    y_coordinates: t.Optional[t.List[float]] = None


class Topic(BaseModel):
    topic_id: TOPIC_ID
    name: str
    lemma_name: t.Optional[str] = None
    term_id: t.List[TERM_ID]
    x_centroid: t.Optional[float] = None
    y_centroid: t.Optional[float] = None
    size: t.Optional[int] = None


class Term(BaseModel):
    term_id: str
    lemma: str
    ent: str
    ngrams: int
    count_terms: int



from bunkatopics.datamodel import DOC_ID
import typing as t
from sklearn.cluster import KMeans
import pandas as pd
from bunkatopics.functions.utils import specificity


def topic_modeling(
    dict_doc_embeddings: t.Dict[DOC_ID, t.Dict[str, int]],
    dict_doc_terms: t.Dict[DOC_ID, t.Dict[str, int]],
    n_clusters=10,
):
    clustering_model = KMeans(n_clusters=n_clusters)

    df_embeddings_2D = pd.DataFrame(dict_doc_embeddings).T
    df_embeddings_2D["topic_number"] = clustering_model.fit(
        df_embeddings_2D
    ).labels_.astype(str)
    df_embeddings_2D["topic_id"] = "bt" + "-" + df_embeddings_2D["topic_number"]
    df_embeddings_2D = df_embeddings_2D.reset_index()
    df_embeddings_2D = df_embeddings_2D.rename(columns={"index": "doc_id"})

    df_topics_docs = df_embeddings_2D.groupby("topic_id").agg(
        size=("doc_id", "count"), x_centroid=("x", "mean"), y_centroid=("y", "mean")
    )
    df_topics_docs = df_topics_docs.reset_index()

    df_terms_topics = pd.DataFrame(dict_doc_terms).T.reset_index()
    df_terms_topics = df_terms_topics.rename(columns={"index": "doc_id"})
    df_terms_topics = pd.merge(df_terms_topics, df_embeddings_2D, on="doc_id")
    df_terms_topics = df_terms_topics.explode("term_id").reset_index(drop=True)

    df_topics_rep = specificity(
        df_terms_topics, X="topic_id", Y="term_id", Z=None, top_n=500
    )

    df_topics_rep = (
        df_topics_rep.groupby("topic_id")["term_id"].apply(list).reset_index()
    )

    df_topics = pd.merge(df_topics_rep, df_topics_docs, on="topic_id")
    df_topics = df_topics.set_index("topic_id")

    df_embeddings_2D = df_embeddings_2D.set_index("doc_id")

    dict_topics = df_topics.to_dict(orient="index")
    dict_docs = df_embeddings_2D.to_dict(orient="index")

    return dict_topics, dict_docs



import matplotlib.pyplot as plt
from sklearn.cluster import KMeans
from kneed import KneeLocator


def compute_knee(data, max_k):
    # Compute sum of squared distances for different values of K
    distortions = []
    for k in range(1, max_k + 1):
        kmeans = KMeans(n_clusters=k)
        kmeans.fit(data)
        distortions.append(
            kmeans.inertia_
        )  # Inertia: Sum of squared distances to the nearest centroid

    # Find the optimal number of clusters using the KneeLocator
    knee = KneeLocator(
        range(1, max_k + 1), distortions, curve="convex", direction="decreasing"
    )

    # Plot the elbow curve with the knee point marked
    plt.plot(range(1, max_k + 1), distortions, marker="o")
    plt.xlabel("Number of clusters (K)")
    plt.ylabel("Sum of squared distances")
    plt.title("Elbow Curve")
    plt.vlines(
        knee.knee,
        plt.ylim()[0],
        plt.ylim()[1],
        linestyles="dashed",
        colors="r",
        label="Knee",
    )
    plt.legend()

    # Print the optimal number of clusters
    print(f"Optimal number of clusters (K): {knee.knee}")
    plt.show()



import typing as t


def remove_overlapping_terms(terms: t.List[str]) -> t.List[str]:
    # 'Harry Style | Harry | Style -> Harry Style'
    seen_words = set()
    filtered_terms = []
    for term in terms:
        if not any(word.lower() in seen_words for word in term.split()):
            filtered_terms.append(term)
            seen_words.update([w.lower() for w in term.split()])
    return filtered_terms



import warnings
from numba.core.errors import NumbaDeprecationWarning

warnings.filterwarnings("ignore", category=NumbaDeprecationWarning)
import pandas as pd
from .bunka_logger import logger
from langchain.embeddings import HuggingFaceInstructEmbeddings
import umap
from sklearn.cluster import KMeans
import warnings
import plotly.graph_objects as go
from .datamodel import Document, Term, Topic, DOC_ID, TOPIC_ID, TERM_ID
from .functions.topic_document import get_top_documents
from .functions.topic_utils import get_topic_repartition
from .visualisation.bourdieu import visualize_bourdieu
from .visualisation.topic_visualization import visualize_topics
from .functions.extract_terms import extract_terms_df
from .functions.topic_representation import remove_overlapping_terms
from .functions.utils import specificity

from .functions.coherence import get_coherence
from .functions.search import vector_search
import uuid
import typing as t

import os

os.environ["TOKENIZERS_PARALLELISM"] = "true"


class Bunka:
    def __init__(self, model_hf=None, language: str = "en_core_web_sm"):
        if model_hf is None:
            model_hf = HuggingFaceInstructEmbeddings(
                model_name="hkunlp/instructor-large",
                embed_instruction="Embed the documents for visualisation of Topic Modeling on a map : ",
                model_kwargs={"device": "cpu"},
                encode_kwargs={"normalize_embeddings": True},
            )
        self.model_hf = model_hf
        self.language = language

    def fit(
        self,
        docs: t.List[DOC_ID],
        multiprocess: bool = True,
    ) -> None:
        df = pd.DataFrame(docs, columns=["content"])
        df["doc_id"] = [str(uuid.uuid4())[:8] for _ in range(len(df))]
        df = df[~df["content"].isna()]
        df = df.reset_index(drop=True)

        docs = [Document(**row) for row in df.to_dict(orient="records")]
        df = pd.DataFrame.from_records([doc.dict() for doc in docs])

        logger.info("Extracting Terms")
        df_terms, df_terms_indexed = extract_terms_df(
            df,
            text_var="content",
            index_var="doc_id",
            ngs=True,
            ents=True,
            ncs=True,
            multiprocess=multiprocess,
            sample_size=100000,
            drop_emoji=True,
            ngrams=(1, 2, 3),
            remove_punctuation=True,
            include_pos=["NOUN", "PROPN", "ADJ"],
            include_types=["PERSON", "ORG"],
            language=self.language,
        )

        df_terms = df_terms.reset_index()
        df_terms = df_terms.rename(columns={"terms_indexed": "term_id"})

        terms = [Term(**row) for row in df_terms.to_dict(orient="records")]

        df_terms_indexed = df_terms_indexed.reset_index()
        df_terms_indexed.columns = ["doc_id", "indexed_terms"]
        indexed_terms_dict = df_terms_indexed.set_index("doc_id")[
            "indexed_terms"
        ].to_dict()

        # add to the docs object
        for doc in docs:
            doc.term_id = indexed_terms_dict.get(doc.doc_id, [])

        sentences = [doc.content for doc in docs]
        ids = [doc.doc_id for doc in docs]

        logger.info("Embedding Documents, this may take few minutes")
        embeddings = self.model_hf.embed_documents(sentences)

        df_embeddings = pd.DataFrame(embeddings)
        df_embeddings.index = ids

        emb_doc_dict = {x: y for x, y in zip(ids, embeddings)}

        for doc in docs:
            doc.embedding = emb_doc_dict.get(doc.doc_id, [])

        logger.info("Reducing Dimensions")
        reducer = umap.UMAP(n_components=2, random_state=42)
        embeddings_2D = reducer.fit_transform(embeddings)

        df_embeddings_2D = pd.DataFrame(embeddings_2D)
        df_embeddings_2D.columns = ["x", "y"]
        df_embeddings_2D["doc_id"] = ids

        xy_dict = df_embeddings_2D.set_index("doc_id")[["x", "y"]].to_dict("index")

        # Update the documents with the x and y values from the DataFrame
        for doc in docs:
            doc.x = xy_dict[doc.doc_id]["x"]
            doc.y = xy_dict[doc.doc_id]["y"]

        self.docs = docs
        self.terms = terms

    def get_topics(self, n_clusters=40, ngrams=[1, 2]):
        clustering_model = KMeans(n_clusters=n_clusters)
        df_embeddings_2D = pd.DataFrame(
            {
                "doc_id": [doc.doc_id for doc in self.docs],
                "x": [doc.x for doc in self.docs],
                "y": [doc.y for doc in self.docs],
            }
        )

        df_embeddings_2D = df_embeddings_2D.set_index("doc_id")

        df_embeddings_2D["topic_number"] = clustering_model.fit(
            df_embeddings_2D
        ).labels_.astype(str)

        df_embeddings_2D["topic_id"] = "bt" + "-" + df_embeddings_2D["topic_number"]

        # insert into the documents
        topic_doc_dict = df_embeddings_2D["topic_id"].to_dict()
        for doc in self.docs:
            doc.topic_id = topic_doc_dict.get(doc.doc_id, [])

        name_lenght = 15
        df_terms = pd.DataFrame.from_records([term.dict() for term in self.terms])
        df_terms = df_terms.sort_values("count_terms", ascending=False)
        df_terms = df_terms.head(1000)
        df_terms = df_terms[df_terms["ngrams"].isin(ngrams)]

        df_terms_indexed = pd.DataFrame.from_records([doc.dict() for doc in self.docs])
        df_terms_indexed = df_terms_indexed[["doc_id", "term_id", "topic_id"]]
        df_terms_indexed = df_terms_indexed.explode("term_id").reset_index(drop=True)

        df_terms_topics = pd.merge(df_terms, df_terms_indexed, on="term_id")

        terms_type = "term_id"
        df_topics_rep = specificity(
            df_terms_topics, X="topic_id", Y=terms_type, Z=None, top_n=500
        )
        df_topics_rep = (
            df_topics_rep.groupby("topic_id")["term_id"].apply(list).reset_index()
        )
        df_topics_rep["name"] = df_topics_rep["term_id"].apply(
            lambda x: x[:name_lenght]
        )
        df_topics_rep["name"] = df_topics_rep["name"].apply(
            lambda x: remove_overlapping_terms(x)
        )
        df_topics_rep["name"] = df_topics_rep["name"].apply(lambda x: " | ".join(x))

        topics = [Topic(**x) for x in df_topics_rep.to_dict(orient="records")]

        df_topics_docs = pd.DataFrame.from_records([doc.dict() for doc in self.docs])
        df_topics_docs = df_topics_docs[["doc_id", "x", "y", "topic_id"]]
        df_topics_docs = df_topics_docs.groupby("topic_id").agg(
            size=("doc_id", "count"), x_centroid=("x", "mean"), y_centroid=("y", "mean")
        )

        topic_dict = df_topics_docs[["size", "x_centroid", "y_centroid"]].to_dict(
            "index"
        )

        # Update the documents with the x and y values from the DataFrame
        for topic in topics:
            topic.size = topic_dict[topic.topic_id]["size"]
            topic.x_centroid = topic_dict[topic.topic_id]["x_centroid"]
            topic.y_centroid = topic_dict[topic.topic_id]["y_centroid"]

        self.topics = topics
        df_topics = pd.DataFrame.from_records([topic.dict() for topic in topics])

        return df_topics

    def fit_transform(self, docs, n_clusters=40):
        self.fit(docs)
        df_topics = self.get_topics(n_clusters=n_clusters)
        return df_topics

    def search(self, user_input: str):
        res = vector_search(self.docs, self.model_hf, user_input=user_input)
        return res

    def get_topic_coherence(self):
        texts = [doc.term_id for doc in self.docs]
        res = get_coherence(self.topics, texts)
        return res

    def get_top_documents(self, top_docs=5) -> pd.DataFrame:
        res = get_top_documents(
            self.docs, self.topics, ranking_terms=20, top_docs=top_docs
        )
        return res

    def get_topic_repartition(self, width=1200, height=800) -> go.Figure:
        fig = get_topic_repartition(self.topics, width=width, height=height)
        return fig

    def visualize_bourdieu(
        self,
        x_left_words=["war"],
        x_right_words=["peace"],
        y_top_words=["men"],
        y_bottom_words=["women"],
        height=1500,
        width=1500,
        clustering=False,
        n_clusters=10,
    ) -> go.Figure:
        fig = visualize_bourdieu(
            self.model_hf,
            self.docs,
            x_left_words=x_left_words,
            x_right_words=x_right_words,
            y_top_words=y_top_words,
            y_bottom_words=y_bottom_words,
            height=height,
            width=width,
            clustering=clustering,
            n_clusters=n_clusters,
        )

        return fig

    def visualize_topics(self, width=1000, height=1000) -> go.Figure:
        fig = visualize_topics(self.docs, self.topics, width=width, height=height)
        return fig



import pandas as pd
from .visu_utils import wrap_by_word
from ..datamodel import Topic, ConvexHullModel, Document
from .convex_hull import get_convex_hull_coord
import numpy as np
import plotly.graph_objects as go
import typing as t


def visualize_topics(
    docs: t.List[Document], topics: t.List[Topic], width=1000, height=1000
):
    df_docs = pd.DataFrame.from_records([doc.dict() for doc in docs])
    try:
        convex_hull_list = []
        for i in set(df_docs["topic_id"]):
            points = df_docs[df_docs["topic_id"] == i][["x", "y"]].values

            # points =  points.reshape(-1, 1)

            x_ch, y_ch = get_convex_hull_coord(points, interpolate_curve=True)
            x_ch = list(x_ch)
            y_ch = list(y_ch)
            convex_hull_list.append(
                ConvexHullModel(topic_id=i, x_coordinates=x_ch, y_coordinates=y_ch)
            )
    except:
        pass

    docs_x = [doc.x for doc in docs]
    docs_y = [doc.y for doc in docs]
    docs_topic_id = [doc.topic_id for doc in docs]
    docs_content = [doc.content for doc in docs]
    docs_content_plotly = [wrap_by_word(x, 10) for x in docs_content]

    topics_x = [topic.x_centroid for topic in topics]
    topics_y = [topic.y_centroid for topic in topics]
    topics_name = [topic.name for topic in topics]
    topics_name_plotly = [wrap_by_word(x, 6) for x in topics_name]
    label_size_ratio = 100

    fig_density = go.Figure(
        go.Histogram2dContour(x=docs_x, y=docs_y, colorscale="delta", showscale=False)
    )

    fig_density.update_traces(contours_coloring="fill", contours_showlabels=False)

    fig_density.update_layout(
        font_size=25,
        width=width,
        height=height,
        margin=dict(
            t=width / 50,
            b=width / 50,
            r=width / 50,
            l=width / 50,
        ),
        title=dict(font=dict(size=width / 40)),
    )

    nk = np.empty(shape=(len(docs_content), 3, 1), dtype="object")
    nk[:, 0] = np.array(docs_topic_id).reshape(-1, 1)
    nk[:, 1] = np.array(docs_content_plotly).reshape(-1, 1)
    # nk[:, 2] = np.array(sizes).reshape(-1, 1)

    # Add points with information
    fig_density.add_trace(
        go.Scatter(
            x=docs_x,
            y=docs_y,
            mode="markers",
            # marker=dict(size=sizes, color=colors),
            # marker=dict(color="#000000"),
            customdata=nk,
            hovertemplate="<br><b>TOPIC</b>: %{customdata[0]}<br>"
            + "<br><b>TEXT</b>: %{customdata[1]}<br>"
            + "<br><b>SCORE</b>: %{customdata[2]}<br>",
        )
    )

    # Add centroids labels
    for x, y, label in zip(topics_x, topics_y, topics_name_plotly):
        fig_density.add_annotation(
            x=x,
            y=y,
            text=label,
            showarrow=True,
            arrowhead=1,
            font=dict(
                family="Courier New, monospace",
                size=width / label_size_ratio,
                # color="#ffffff",
                color="blue",
            ),
            bordercolor="#c7c7c7",
            borderwidth=width / 1000,
            borderpad=width / 500,
            # bgcolor="#ff7f0e",
            bgcolor="white",
            opacity=1,
            arrowcolor="#ff7f0e",
        )

    try:
        for convex_hull in convex_hull_list:
            # Create a Scatter plot with the convex hull coordinates
            trace = go.Scatter(
                x=convex_hull.x_coordinates,
                y=convex_hull.y_coordinates,  # Assuming y=0 for simplicity
                mode="lines",
                name="Convex Hull",
                line=dict(color="grey"),
            )

            fig_density.add_trace(trace)
    except:
        pass

    logo_image = "Bunka Communication - obordaibng (2).png"
    logo_width = 50
    logo_height = 50

    # Add the logo image
    fig_density.add_layout_image(
        source=logo_image,
        xref="paper",
        yref="paper",
        x=1,
        y=1,
        sizex=logo_width,
        sizey=logo_height,
        xanchor="right",
        yanchor="top",
        opacity=1,
    )

    fig_density.update_layout(showlegend=False)
    fig_density.update_xaxes(showgrid=False, showticklabels=False, zeroline=False)
    fig_density.update_yaxes(showgrid=False, showticklabels=False, zeroline=False)
    fig_density.update_yaxes(showticklabels=False)

    return fig_density



import numpy as np
from gensim.models import CoherenceModel
from gensim.corpora.dictionary import Dictionary
import typing as t
from ..datamodel import TERM_ID


def get_coherence(
    topics: t.List[t.List[TERM_ID]],
    texts: t.List[t.List[TERM_ID]],
    coherence_type="c_v",
    topic_terms_n: int = 10,
):
    word2id = Dictionary(texts)
    topics_terms = [topic.term_id[:topic_terms_n] for topic in topics]
    cm = CoherenceModel(
        topics=topics_terms, texts=texts, coherence=coherence_type, dictionary=word2id
    )
    coherence_per_topic = cm.get_coherence_per_topic()
    coherence_mean = np.mean(coherence_per_topic)

    return coherence_mean



import plotly.express as px
import pandas as pd
import typing as t
from bunkatopics.visualisation.visu_utils import wrap_by_word
from ..datamodel import Topic


def get_topic_repartition(topics: t.List[Topic], width=1200, height=800):
    df_topics = pd.DataFrame.from_records([topic.dict() for topic in topics])
    df_topics["name"] = df_topics["name"].apply(lambda x: wrap_by_word(x, 8))
    df_topics = df_topics.sort_values("size")
    fig = px.bar(
        df_topics,
        y="name",
        x="size",
        width=width,
        height=height,
        title="Topic Size",
        template="simple_white",
    )
    return fig



import typing as t
from ..datamodel import Document, Topic
import pandas as pd


def get_top_documents(
    docs: t.List[Document], topics: t.List[Topic], ranking_terms=20, top_docs=5
) -> pd.DataFrame:
    df_docs = pd.DataFrame.from_records([doc.dict() for doc in docs])
    df_docs = df_docs[["doc_id", "topic_id", "term_id"]]
    df_docs = df_docs.explode("term_id").reset_index(drop=True)

    df_topics = pd.DataFrame.from_records([topic.dict() for topic in topics])
    df_topics["term_id"] = df_topics["term_id"].apply(lambda x: x[:ranking_terms])
    df_topics = df_topics[["topic_id", "term_id"]]
    df_topics = df_topics.explode("term_id").reset_index(drop=True)

    df_rank = pd.merge(df_docs, df_topics, on=["topic_id", "term_id"])
    df_rank = (
        df_rank.groupby(["topic_id", "doc_id"])["term_id"]
        .count()
        .rename("count_topic_terms")
        .reset_index()
    )
    df_rank = df_rank.sort_values(
        ["topic_id", "count_topic_terms"], ascending=(True, False)
    ).reset_index(drop=True)
    df_rank = df_rank.groupby("topic_id").head(5).reset_index(drop=True)

    df_docs = pd.DataFrame.from_records([doc.dict() for doc in docs])
    df_docs = df_docs[["doc_id", "content"]]

    df_rank = pd.merge(df_rank, df_docs, on="doc_id")
    df_topics = pd.DataFrame.from_records([topic.dict() for topic in topics])
    df_topics = df_topics[["topic_id", "name"]]

    df_rank = pd.merge(df_rank, df_topics, on="topic_id")

    return df_rank



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bunkatopics',
 'bunkatopics.bunka_logger',
 'bunkatopics.functions',
 'bunkatopics.visualisation']

package_data = \
{'': ['*']}

install_requires = \
['black>=23.3.0,<24.0.0',
 'gensim>=4.3.1,<5.0.0',
 'hdbscan>=0.8.29,<0.9.0',
 'instructorembedding>=1.0.1,<2.0.0',
 'ipywidgets>=8.0.6,<9.0.0',
 'jupyterlab>=4.0.1,<5.0.0',
 'kneed>=0.8.3,<0.9.0',
 'langchain>=0.0.188,<0.0.189',
 'loguru>=0.7.0,<0.8.0',
 'matplotlib>=3.7.1,<4.0.0',
 'numba>=0.57.0,<0.58.0',
 'pandas>=2.0.2,<3.0.0',
 'plotly>=5.14.1,<6.0.0',
 'pytest>=7.3.2,<8.0.0',
 'sentence-transformers>=2.2.2,<3.0.0',
 'statsmodels>=0.14.0,<0.15.0',
 'textacy>=0.13.0,<0.14.0',
 'umap-learn>=0.5.3,<0.6.0']

setup_kwargs = {
    'name': 'bunkatopics',
    'version': '0.37',
    'description': 'Topic Modeling using Transformers and advanced visualization',
    'long_description': '[![PyPI - Python](https://img.shields.io/badge/python-v3.9+-blue.svg)](https://pypi.org/project/bertopic/)\n[![PyPI - PyPi](https://img.shields.io/pypi/v/bunkatopics)](https://pypi.org/project/bunkatopics/)\n\n# Bunkatopics\n\n<img src="images/logo.png" width="35%" height="35%" align="right" />\n\nBunkatopics is a Topic Modeling Visualisation Method that leverages Transformers from HuggingFace through langchain. It is built with the same philosophy as [BERTopic](https://github.com/MaartenGr/BERTopic) but goes deeper in the visualization to help users grasp quickly and intuitively the content of thousands of text.\nIt also allows for a supervised visual representation by letting the user create continnums with natural language.\n\n## Installation\n\nFirst, create a new virtual environment using pyenv\n\n```bash\npyenv virtualenv 3.9 bunkatopics_env\n```\n\nActivate the environment\n\n```bash\npyenv activate bunkatopics_env\n```\n\nThen Install the Bunkatopics package:\n\n```bash\npip install bunkatopics\n```\n\nInstall the spacy tokenizer model for english:\n\n```bash\npython -m spacy download en_core_web_sm\n```\n\n## Contributing\n\nAny contribution is more than welcome\n\n```bash\npip install poetry\ngit clone https://github.com/charlesdedampierre/BunkaTopics.git\ncd BunkaTopics\n\n# Create the environment from the .lock file. \npoetry install # This will install all packages in the .lock file inside a virtual environmnet\n\n# Start the environment\npoetry shell\n```\n\n## Getting Started\n\n| Name  | Link  |\n|---|---|\n| Visual Topic Modeling With Bunkatopics  | [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/drive/1DtPrI82TYepWLoc4RwuQnOqMJb0eWT_t?usp=sharing)  |\n\n## Quick Start\n\nWe start by extracting topics from the well-known 20 newsgroups dataset containing English documents:\n\n```python\nfrom bunkatopics import bunkatopics\nfrom sklearn.datasets import fetch_20newsgroups\nimport random\n \nfull_docs = fetch_20newsgroups(subset=\'all\',  remove=(\'headers\', \'footers\', \'quotes\'))[\'data\']\nfull_docs_random = random.sample(full_docs, 1000)\n\n```\n\nYou can the load any model from langchain. Some of them might be large, please check the langchain [documentation](https://python.langchain.com/en/latest/reference/modules/embeddings.html)\n\nIf you want to start with a small model:\n\n```python\nfrom langchain.embeddings import HuggingFaceEmbeddings\n\nembedding_model = HuggingFaceEmbeddings(model_name="all-MiniLM-L6-v2")\n\nbunka = Bunka(model_hf=embedding_model)\n\nbunka.fit(full_docs)\ndf_topics = bunka.get_topics(n_clusters = 20)\n```\n\nIf you want a bigger LLM Like [Instructor](https://github.com/HKUNLP/instructor-embedding)\n\n```python\nfrom langchain.embeddings import HuggingFaceInstructEmbeddings\n\nembedding_model = HuggingFaceInstructEmbeddings(model_name="hkunlp/instructor-large",\n                                                embed_instruction="Embed the documents for visualisation of Topic Modeling on a map : ")\n\nbunka = Bunka(model_hf=embedding_model)\n\nbunka.fit(full_docs)\ndf_topics = bunka.get_topics(n_clusters = 20)\n```\n\nThen, we can visualize\n\n```python\ntopic_fig = bunka.visualize_topics( width=800, height=800)\ntopic_fig\n...\n```\n\nThe map display the different texts on a 2-Dimensional unsupervised scale. Every region of the map is a topic described by its most specific terms.\n\n<img src="images/newsmap.png" width="70%" height="70%" align="center" />\n\n```python\n\nbourdieu_fig = bunka.visualize_bourdieu(x_left_words=["past"],\n                                        x_right_words=["future", "futuristic"],\n                                        y_top_words=["politics", "Government"],\n                                        y_bottom_words=["cultural phenomenons"],\n                                        height=2000,\n                                        width=2000)\n```  \n\nThe power of this visualisation is to constrain the axis by creating continuums and looking how the data distribute over these continuums. The inspiration is coming from the French sociologist Bourdieu, who projected items on [2 Dimensional maps](https://www.politika.io/en/notice/multiple-correspondence-analysis).\n\n<img src="images/bourdieu.png" width="70%" height="70%" align="center" />\n\n## Multilanguage\n\nThe package use Spacy to extract meaningfull terms for the topic represenation.\n\nIf you wish to change language to french, first, download the corresponding spacy model:\n\n```bash\npython -m spacy download fr_core_news_lg\n```\n\n```python\nfrom langchain.embeddings import HuggingFaceEmbeddings\n\nembedding_model = HuggingFaceEmbeddings(model_name="distiluse-base-multilingual-cased-v2")\n\nbunka = Bunka(model_hf=embedding_model, language = \'fr_core_news_lg\')\n\nbunka.fit(full_docs)\ndf_topics = bunka.get_topics(n_clusters = 20)\n```  \n\n## Functionality\n\nHere are all the things you can do with Bunkatopics\n\n### Common\n\nBelow, you will find an overview of common functions in BERTopic.\n\n| Method | Code  |\n|-----------------------|---|\n| Fit the model    |  `.fit(docs)` |\n| Fit the model and get the topics  |  `.fit_transform(docs)` |\n| Acces the topics   | `.get_topics(n_clusters=10)`  |\n| Access the top documents per topic    |  `.get_top_documents()` |\n| Access the distribution of topics   |  `.get_topic_repartition()` |\n| Visualize the topics on a Map |  `.visualize_topics()` |\n| Visualize the topics on Natural Language Supervised axis | `.visualize_bourdieu()` |\n| Access the Coherence of Topics |  `.get_topic_coherence()` |\n| Get the closest documents to your search | `.search(\'politics\')` |\n\n### Attributes\n\nYou can access several attributes\n\n| Attribute | Description |\n|------------------------|---------------------------------------------------------------------------------------------|\n| `.docs`               | The documents stores as a Document pydantic model |\n| `.topics` | The Topics stored as a Topic pydantic model. |\n',
    'author': 'Charles De Dampierre',
    'author_email': 'charlesdedampierre@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
